/* 
 * $Id: benchread.c,v 1.6 1999/06/17 16:43:58 fswasey Exp $
 *
 * Name: benchread
 *
 * Function: benchmark readers of a news server
 *
 * Documentation:  process this documentation with the command:
 *	pod2man --section 8 benchread.c >benchread.8

=head1 NAME

benchread - a program to help benchmark news reader performance.

=head1 SYNOPSIS

B<benchread> [ B<-v> ... ] I<control-file>

=head1 DESCRIPTION

B<benchread> generates a set of reader transactions against the news server
specified in the I<control-file> and prints a report on STDOUT. 
Transaction data is available via use of the C<-v> option and is directed
to STDERR.

=head1 OPTIONS

=over 8

=item -v

Increment the level of output to STDERR.

The default level of output is none.
Everytime C<-v> is specified, the level is incremented by one (1).
Valid levels are:

=over 8

=item 0

No transaction process information is printed.

=item 1

Information telling you the command sent,
the first line of output returned,
and the amount of sleep time between commands
is sent to STDERR.

=item 2

All the information from the previous level
is printed and the entire return message from
the server is sent to STDERR.

=back

=back

=head1 FILES

=head2 control-file

The I<control-file> consists of a series of header: value pairs.
Valid header values are:

=over 8

=item server

The name of the server to benchmark.

=item port

The port the server is listening on.
Defaults to 119.

=item groups

The comma delimited list of news groups to read.
Spaces are not allowed between the group names.
Normal line continuation using the \ (backslash) is permitted.

=item group_count
=item group_initial
=item group_interval

If the B<groups> header is not specified,
then B<group_count>, B<group_initial>, and B<group_interval> are required.
B<group_count> specifies the number of groups to select for reading.

B<group_initial> specifies the count of lines returned from a I<list active>
command to use as the first group.

B<group_interval> specifies the maximum number of groups to skip between 
selections.

The number (I<n>) specified for B<group_count>, B<group_initial>,
and B<group_interval> is used differently depending on whether it is positive
or negative.
If the I<n> is positive,
a random number is picked between 0 and 2^I<n>.
If the I<n> is negative, abs(I<n>) is the number.

=item group_delay

Specifies the boundary of the number of seconds to sleep before all
but the first group commands.

If the number I<n> is positive, a random number is selected
between 0 and 2^I<n>.
If the I<n> is negative, abs(I<n>) is the number selected.

Each other than the first group command is to be issued a random number of
seconds between 0 and the number selected is waited.

=item xover_delay

Specifies the boundary of the number of seconds to sleep before all 
xover commands.

The value is determined and used as the B<group_delay> value is above.

=item art_delay

Specifies the boundary of the number of seconds to sleep before all 
article commands.

The value is determined and used as the B<group_delay> value is above.

=item seed

Specifies the random number seed.
This is not a required header and defaults to the process id if not specified.
It is included so you can force multiple runs to be identical
(assuming nothing has changed on the server you are testing).

=back

=head1 EXAMPLES

	server:	news.foo.com
	port:	119
	group_count:	-10
	group_initial:	10
	group_interval:	6
	group_delay:	5
	xover_delay:	0
	art_delay:	5

The above I<control-file> will read 10 groups from news.foo.com on port 119.
The first group will be randomly selected between lines 0 and 1024 returned
from the B<list active> command.
Subsequent groups to fill out the 10, will be selected randomly between 0
and 64 lines further on.
Prior to issuing a B<group> command, a random sleep between 0 and 32 seconds
will be generated.
A random sleep between 0 and 1 second will be generated prior to each 
B<xover> command.
Each B<article> command will be preceeded by a random sleep of 0 to 32 seconds.

When the process is complete, a report will be generated on STDOUT which
looks something like the following:

	Control file:
	server:		news.foo.com
	port:		119
	groups:
	group_count:	-10
	group_initial:	10
	group_interval:	6
	group_delay:	5
	xover_delay:	0
	art_delay:	5

	Generated Entries:
	seed:	43718
	groups:	alt.music.beastie-boys,alt.music.canada,alt.music.mor\
		rissey,alt.music.producer,alt.music.smash-pumpkins,al\
		t.org.food-not-bombs,alt.org.team-os2,alt.pagan.conta\
		cts,alt.personals.tall,alt.private.investigator

	Started at: Tue Jun 15 15:01:10 1999
	Ended at:   Thu Jun 17 10:45:51 1999

	Groups:		10 / 10
	Articles:	9629 / 9630

		WALLCLOCK	   CPU		Percent
	group	   5.35s	   0.01s	   0.22%
	xover	 643.61s	   0.33s	  26.79%
	article	1753.56s	   7.55s	  72.99%
		=======		=======		=======
	TOTAL:	2402.52s	   7.89s	 100.00%

=head1 BUGS

The B<clients> header has not been implemented yet to allow you
to run multiple reader loads on your server easily.

=head1 AUTHOR

Frank Swasey <swasey@together.net>

=cut

 *
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<time.h>
#include	<unistd.h>
#include	<string.h>
#include	<sys/times.h>
#include	<sys/types.h>
#include	<netdb.h>
#include	<netinet/in.h>
#include	<sys/socket.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<math.h>
#include	<errno.h>
#include	<libgen.h>

#if !defined(TRUE)
#define	TRUE	1
#define	FALSE	0
#endif

#define	MAXLINE	4096
struct times {
	float		wallclock;
	float		cpu;
	float		percent;
};
	
static void	pr_times(char *, struct times *);
static void	do_cmd(int, char *, struct times *);
static void	get_line(int, char *, int);
static void	Receive_Line(int, char *, int);
static int	NNTP_code(char *);
static void	process_msg(int, char *);
static void	parse_file(char *);
static void	read_file_line(FILE *, char *, int);
static void	C_groups(char *);
static void	C_number(int *, char *);
static void	C_string(char *, int, char *);
static char *	find_data(char *);
static void	dump_groups(char **);
static void	get_group(char *, char **, char);

/* Nasty global variables.... */
int		low, high;	/* low and high article numbers */
char 		Server[1024] = "";
int		port = 119;
int		group_initial;
int		group_interval;
int		group_count;
int		group_delay;
int		xover_delay;
int		article_delay;
int		groups_read = 0;
int		group_commands = 0;
int		articles_read = 0;
int		article_commands = 0;
char **		group_names = NULL;
char **		article_numbers = NULL;
int		Verbose = 0;
int		seed = 0;

char *		pname;

int
main(int argc, char *argv[])
{
	int			i;	/* General Index */
	struct times		times_group;
	struct times		times_xover;
	struct times		times_article;
	struct times		times_total;
	time_t			s, e;
	struct sockaddr_in	sin;
	struct hostent *	hp;
	int			sd;
	unsigned int		sleep_amt;
	char			msg[MAXLINE];
	char **			q;
	extern char *		optarg;
	extern int		optind, opterr, optopt;

	/* get ready for those errors, that we hope won't happen */
	pname = basename(argv[0]);
	/* Parse command line... */
	while ( (i=getopt(argc,argv,"v")) != -1 ) {
		switch (i) {
		case 'v':
			++Verbose;
			break;
		default:
			fprintf(stderr, "Syntax: %s [-v] control_file\n",
				pname);
			exit(1);
			break;
		}
	}
	
	/* There better be something left -- I'm good, but I'm not good
	 * enough to create a control file out of thin air!
	 */
	if (optind >= argc || optind < (argc - 1)) {
		fprintf(stderr, "Syntax: %s [-v] control_file\n", pname);
		exit(1);
	}
	else {
		fprintf(stdout, "Control file:\n");
		parse_file(argv[optind]);
	}
	
	/* Initialize */
	times_group.wallclock = times_group.cpu = 
		times_group.percent = 0.0;
	times_xover.wallclock = times_xover.cpu = 
		times_xover.percent = 0.0;
	times_article.wallclock = times_article.cpu = 
		times_article.percent = 0.0;
	times_total.wallclock = times_total.cpu = 
		times_total.percent = 0.0;
	
	if (seed == 0 || group_names == NULL)
		fprintf(stdout, "Generated Entries:\n");

	if (seed == 0) {
		seed = getpid();
		fprintf(stdout, "seed:\t%d\n", seed);
	}
	srand(seed);
	if (group_names == NULL) {
		if (group_count < 0)
			group_count = abs(group_count);
		else
			group_count = rand() % (int) (pow(2,group_count)+1);
		if (group_initial < 0)
			group_initial = abs(group_initial);
		else
			group_initial = rand() % (int) (pow(2,group_initial)+1);
		if (group_interval < 0)
			group_interval = abs(group_interval);
		else
			group_interval = (int) (pow(2,group_interval));
	}
	if (group_delay < 0) 
		group_delay = abs(group_delay);
	else
		group_delay = (int)pow(2,group_delay);
	if (xover_delay < 0)
		xover_delay = abs(xover_delay);
	else
		xover_delay = (int)pow(2,xover_delay);
	if (article_delay < 0)
		article_delay = abs(article_delay);
	else
		article_delay = (int)pow(2,article_delay);
		
	if ((hp = gethostbyname(Server)) == 0) {
		fprintf(stderr, "gethostbyname error: %s\n", strerror(errno));
		exit(1);
	}

	bzero(&sin, sizeof(sin));
	bcopy(hp->h_addr, &sin.sin_addr, hp->h_length);
	sin.sin_family = hp->h_addrtype;
	sin.sin_port = htons(port);

	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		fprintf(stderr, "socket error: %s\n", strerror(errno));
		exit(1);
	}
	
	/* Connect to the master server */
	if (connect(sd, (struct sockaddr *)&sin, sizeof(sin)) == -1) {
		fprintf(stderr, "connect error: %s\n", strerror(errno));
		exit(1);
	}
	
	/* Get that header... */
	get_line(sd, msg, sizeof(msg));
	if (Verbose>0) fprintf(stderr, "<--:%s\n", msg);

	/* Do we need to determine a list of groups? */
	if (group_names == NULL) {
		do_cmd(sd, "list active", &times_total);
		dump_groups(group_names);
		times_total.wallclock = times_total.cpu = 
			times_total.percent = 0.0;
	}

	/* Start the clock */
	if ( time(&s) == -1 ) {
		fprintf(stderr, "time error");
		exit(1);
	}

	/* Have the conversation */
	while (*group_names) {
		if (group_commands > 0) {
			sleep_amt = rand() % (group_delay + 1);
			if (Verbose>0)
				fprintf(stderr, "Sleeping for %d %s\n",
					sleep_amt,
					sleep_amt != 1 ? "seconds" : "second");
			while (sleep_amt != 0)
				sleep_amt = sleep(sleep_amt);
		}
		sprintf(msg, "group %s", *group_names++);
		++group_commands;
		do_cmd(sd, msg, &times_group);
		if (low < high || low > 0) {
			sleep_amt = rand() % (xover_delay + 1);
			if (Verbose>0)
				fprintf(stderr, "Sleeping for %d %s\n",
					sleep_amt,
					sleep_amt != 1 ? "seconds" : "second");
			while (sleep_amt != 0)
				sleep_amt = sleep(sleep_amt);
			sprintf(msg, "xover %d-%d", low, high);
			do_cmd(sd, msg, &times_xover);
			q = article_numbers;
			while (*q) {
				sleep_amt = rand() % (article_delay + 1);
				if (Verbose>0)
					fprintf(stderr, "Sleeping for %d %s\n",
						sleep_amt,
						sleep_amt != 1 ?
							 "seconds" : "second");
				while (sleep_amt != 0)
					sleep_amt = sleep(sleep_amt);
				sprintf(msg, "article %s", *q);
				free(*q);
				*q++ = NULL;
				++article_commands;
				do_cmd(sd, msg, &times_article);
			}
			if (article_numbers) {
				free(article_numbers);
				article_numbers = NULL;
			}
		}
	}
	do_cmd(sd, "quit", &times_group);

	if ( time(&e) == -1 ) {
		fprintf(stderr, "time error: %s\n", strerror(errno));
		exit(1);
	}

	close(sd);

	fprintf(stdout, "\nStarted at: %s", ctime(&s));
	fprintf(stdout, "Ended at:   %s", ctime(&e));
	fprintf(stdout, "\nGroups:\t\t%d / %d\nArticles:\t%d / %d\n", 
		groups_read, group_commands,
		articles_read, article_commands);
	/* Calculate totals and percentages */
	/* Total Real Time (wall clock) */
	times_total.wallclock = times_group.wallclock + 
		times_xover.wallclock + times_article.wallclock;
	/* Total CPU Time */
	times_total.cpu = times_group.cpu + 
		times_xover.cpu + times_article.cpu;
	/* Percentages of Real Time */
	times_group.percent =
		(times_group.wallclock*100.0)/times_total.wallclock;
	times_xover.percent =
		(times_xover.wallclock*100.0)/times_total.wallclock;
	times_article.percent =
		(times_article.wallclock*100.0)/times_total.wallclock;
	/* Total Percent Real Time */
	times_total.percent = times_group.percent + 
		times_xover.percent + times_article.percent;
	
	/* Print the numbers... */
	printf("\n\tWALLCLOCK\t   CPU\t\tPercent\n");
	pr_times("group",  &times_group);
	pr_times("xover",  &times_xover);
	pr_times("article",&times_article);
	printf("\t=======\t\t=======\t\t=======\n");
	pr_times("TOTAL:", &times_total);
}

static void
do_cmd(int sd, char *cmd, struct times *x)
{
	struct tms	tmsstart, tmsend;
	clock_t		start, end;
	char		msg[MAXLINE];
	int		i;

	if (Verbose>0) fprintf(stderr, ":-->%s\n", cmd);

	sprintf(msg, "%s\r\n", cmd);

	if ( (start = times(&tmsstart)) == -1) {	/* starting values */
		fprintf(stderr, "times error: %s\n", strerror(errno));
		exit(1);
	}

	if (send(sd, msg, strlen(msg), 0) == -1) {
		fprintf(stderr, "system() error: %s\n", strerror(errno));
		exit(1);
	}

	get_line(sd, msg, sizeof(msg));
	if (Verbose>0) fprintf(stderr, "<--:%s\n", msg);

	process_msg(sd, msg);

	if ( (end = times(&tmsend)) == -1) {		/* ending values */
		fprintf(stderr, "times error: %s\n", strerror(errno));
		exit(1);
	}

	x->wallclock += (float) ( end - start );	/* add up real time */
	x->cpu += (float) ((tmsend.tms_utime - tmsstart.tms_utime) +
			(tmsend.tms_stime - tmsstart.tms_stime));
}

static void
pr_times(char *header, struct times *x)
{
	static long		clktck = 0;

	if (clktck == 0)	/* fetch clock ticks per second first time */
		if ( (clktck = sysconf(_SC_CLK_TCK)) < 0) {
			fprintf(stderr, "sysconf error: %s\n", strerror(errno));
			exit(1);
		}
	fprintf(stdout, "%s\t%7.2fs\t%7.2fs\t%7.2f%%\n",
			header,
			x->wallclock / (double) clktck,
			x->cpu / (double) clktck,
			x->percent);
}

static void
get_line(int sd, char *msg, int len)
{
	static char	line[MAXLINE];
	int		i;
	static char *	p = &line[MAXLINE]+1;

	i = 0;
	while ( i < len-1 ) {
		if (p > &line[MAXLINE] || *p == '\0') {
			Receive_Line(sd, line, MAXLINE);
			p = &line[0];
		}
		if (*p == '\r') {
			p++;
			break;
		}
		
		if (*p == '\n') {
			p++;
		}
		else {

			msg[i++] = *p++;
		}
	}
	if ( i < len )
		msg[i] = '\0';
}
		
		
static void
Receive_Line(int sd, char *msg, int len)
{
	int		i;

	i = 0;
	while ( i == 0 ) {
		if ( ( i=recv(sd, msg, len, 0)) == -1) {
			fprintf(stderr, "recv error: %s\n", strerror(errno));
			exit(1);
		}
	}
	msg[i] = '\0';
}

static int
NNTP_code(char *txt)
{
	int	i=0;

	while ( *txt && (*txt >= '0' && *txt <= '9') ) {
		i = (i*10) + (*txt++ - '0');
	}
	return i;
}

static void
process_msg(int sd, char *msg)
{
	char	nmsg[MAXLINE];
	int	i, j;
	char **	q;
	int	nntpcode;

	nntpcode = NNTP_code(msg);
	switch (nntpcode) {
	case 220:	/* n <a> article retrieved - head and body follow */
	case 221:	/* n <a> article retrieved - head follows */
	case 222:	/* n <a> article retrieved - body follows */
		articles_read++;
		/* No BREAK here, fall through is intentional */

	case 100:	/* help text follows */
		nmsg[0] = '\0';
		nmsg[1] = '\0';
		while ( !(nmsg[0] == '.' && nmsg[1] == '\0') ) {
			get_line(sd, nmsg, sizeof(nmsg));
			if (Verbose>1) fprintf(stderr, "<--:%s\n", nmsg);
		}
		break;

	case 224:	/* data follows -- response to xover */
		if (article_numbers) {
			fprintf(stderr, "article_numbers not null at alloc time");
			exit(1);
		}
		j = high - low + 1;	/* count of articles to expect */
#if defined(DEBUG)
		fprintf(stderr, "expecting %d articles [%d-%d]\n",
			j, low, high);
#endif
		if ((article_numbers=malloc((j+1)*sizeof(char *))) == NULL) {
			fprintf(stderr,
				"malloc for article numbers failed: %s\n",
				strerror(errno));
			exit(1);
		}
		q = article_numbers;
		*q = NULL;
		nmsg[0] = '\0';
		nmsg[1] = '\0';
		while ( !(nmsg[0] == '.' && nmsg[1] == '\0') ) {
			get_line(sd, nmsg, sizeof(nmsg));
#if defined(DEBUG)
			fprintf(stderr, "<--:%s\n", nmsg);
#else
			if (Verbose>1) fprintf(stderr, "<--:%s\n", nmsg);
#endif
			if (nmsg[0] != '.') { /* because of xover, . means end*/
				if (j-- == 0) {
					fprintf(stderr, "Too many articles from xover.");
					exit(1);
				}
				get_group(nmsg, q, '\t'); /*get article number*/
				q++;
				*q = NULL;
#if defined(DEBUG)
				fprintf(stderr, "%d articles left to get...\n",
					j);
#endif
			}
		}
		break;

	case 215:	/* list of newsgroups follows */
	case 231:	/* list of new newsgroups follows */
		/* assumption is that we did a "list active" command and
		 * are trying to build a list of groups to read...
		 */
		group_names = malloc((group_count + 1)*sizeof(char *));
		if (group_names == NULL) {
			fprintf(stderr, "unable to malloc %d bytes: %s\n",
				((group_count + 1)*sizeof(char *)),
				strerror(errno));
			exit(1);
		}
		q = group_names;
		i = 0;
		j = group_initial;
		nmsg[0] = '\0';
		nmsg[1] = '\0';
		while ( !(nmsg[0] == '.' && nmsg[1] == '\0') ) {
			get_line(sd, nmsg, sizeof(nmsg));
			if (Verbose>1) fprintf(stderr, "<--:%s\n", nmsg);
			if (j == i) {
				get_group(nmsg, q, ' ');
				q++;
				*q = NULL;
				group_count--;
				if (group_count > 0) 
					j += rand() % (group_interval + 1);
				else
					j = i-1;
			}
			i++;
		}
		break;

	case 211:	/* n f l s group selected */
		while (*msg && *msg != ' ') msg++; /* skip 211 */
		if (*msg) msg++;
		while (*msg && *msg != ' ') msg++; /* skip n (number of arts) */
		if (*msg) msg++;
		low = NNTP_code(msg);
		while (*msg && *msg != ' ') msg++; /* skip f (low) */
		if (*msg) msg++;
		high = NNTP_code(msg);
		groups_read++;	/* increment number of groups read */
		break;
	case 411:	/* bad group selected */
		low = high = 0;
		break;
	
	case 423:	/* No such article number in this group */
	case 430:	/* No such article found */
	case 205:	/* Adios */
		break;

	default:
		fprintf(stderr, "NNTP Code (%d) unexpected -- aborting", nntpcode);
		exit(1);
		break;
	}
}

static void
parse_file(char *filename)
{
	FILE *		fd;
	char		line[MAXLINE];
	int		i, j, contline;

	if ( (fd = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "%s Cannot open file '%s': %s\n",
			pname, filename, strerror(errno));
		exit(1);
	}
	
	/* Read and parse the file... */
	while (!feof(fd)) {
		read_file_line(fd, line, MAXLINE);
		if (strncmp("server:", line, 7) == 0) {
			C_string(Server, sizeof(Server), line);
		}
		else if (strncmp("port:", line, 5) == 0) {
			C_number(&port, line);
		}
		else if (strncmp("groups:", line, 7) == 0) {
			if (strlen(line) > 7)
				C_groups(line);
		}
		else if (strncmp("group_count:", line, 12) == 0) {
			C_number(&group_count, line);
		}
		else if (strncmp("group_initial:", line, 14) == 0) {
			C_number(&group_initial, line);
		}
		else if (strncmp("group_interval:", line, 15) == 0) {
			C_number(&group_interval, line);
		}
		else if (strncmp("group_delay:", line, 12) == 0) {
			C_number(&group_delay, line);
		}
		else if (strncmp("xover_delay:", line, 12) == 0) {
			C_number(&xover_delay, line);
		}
		else if (strncmp("art_delay:", line, 10) == 0) {
			C_number(&article_delay, line);
		}
		else if (strncmp("seed:", line, 5) == 0) {
			C_number(&seed, line);
		}
		else {
			fprintf(stderr, "Unknown statement: '%s'", line);
			exit(1);
		}
	}

	/* Validate the file... */
	if (group_names != NULL) {
		if (group_initial != 0 ||
			group_count != 0 ||
			group_interval != 0 ) {
			fprintf(stderr, "Illegal use of groups and group_count");
			exit(1);
		}
	}
	else if (group_initial == 0 ||
			group_count == 0 ||
			group_interval == 0 ) {
		fprintf(stderr, "Must specify either groups: or group_initial, group_count, and group_interval");
		exit(1);
	}
}

static void
read_file_line(FILE *fd, char *line, int len)
{
	int i, j, contline;

	i = 0;
	contline = FALSE;
	while (1 == 1) {
		if ((j = fgetc(fd)) == EOF) {
			if (i != 0) {
				fprintf(stderr, "Unexpected end of file: %s\n",
					strerror(errno));
				exit(1);
			}
			else
				j = '\n';
		}
		fputc(j, stdout);
		switch (j) {
		case '\n':
			if ( i==0 || line[i-1] != '\\' )
					return;
			else {
				line[--i] = '\0';
				contline = TRUE;
			}
			break;
		case ' ':
		case '\t':
			if (contline) {
				break;
			}
		default:
			line[i++] = j;
			line[i] = '\0';
			break;
		}
	}
}

static void
C_groups(char *line)
{
	int	i = 1;
	char *	p;
	char **	q;

	line = find_data(line);

	/* determine how many groups... */
	p = line;
	while (*p) {
		if (*p++ == ',') i++;
	}

	group_names = malloc( (i+1) * sizeof(char *));
	if (group_names == NULL) {
		fprintf(stderr, "unable to malloc %d bytes: %s\n",
			(i+1)*sizeof(char *), strerror(errno));
		exit(1);
	}

	/* copy groups */
	q = group_names;
	p = line;
	while (*p) {
		if (*p == ',') {
			*p='\0';
			if ( (*q=malloc(strlen(line)+1)) == NULL) {
				fprintf(stderr,
					"unable to malloc %d bytes: %s\n",
					strlen(line)+1, strerror(errno));
				exit(1);
			}
			strcpy(*q, line);
			line = p+1;
			q++;
		}
		p++;
	}
	if (strlen(line) != 0) {
		if ( (*q=malloc(strlen(line)+1)) == NULL) {
			fprintf(stderr, "unable to malloc %d bytes: %s\n",
				strlen(line)+1, strerror(errno));
			exit(1);
		}
		strcpy(*q, line);
		q++;
	}
	*q = NULL;
}

static void
C_number(int *n, char *line)
{
	line = find_data(line);
	*n = atoi(line);
}

static void
C_string(char *s, int l_s, char *line)
{
	line = find_data(line);
	strncpy(s, line, l_s);
	s[l_s-1] = '\0';	/* Make sure it's null terminated */
}

static char *
find_data(char *s)
{
	char *p = s;

	while (*p != ':')	/* skip header */
		p++;

	p++;	/* skip that colon */

	while (*p == ' ' || *p == '\t')	/* skip the spaces/tabs */
		p++;

	return p;
}

static void
dump_groups(char **x)
{
	int i = 0;
	int lpos;
	char *q;

	if (x) {
		fprintf(stdout, "groups:\t");
		lpos = 9;
		while (*x) {
			q = *x;
			if (i != 0) {
				fputc(',', stdout);
				lpos++;
			}
			while (*q) {
				if (lpos >= 75) {
					fputs("\\\n\t", stdout);
					lpos=9;
				}
				fputc(*q++, stdout);
				lpos++;
			}
			x++;
			i++;
		}
		fputc('\n', stdout);
	}
}

static void
get_group(char *line, char **q, char sep)
{
	char *p;

	p = line;
	while (*p) {
		if (*p == sep ) {
#if defined(DEBUG)
			fprintf(stderr, "Found that first '%c' character\n",
				sep);
#endif
			*p = '\0';
#if defined(DEBUG)
			fprintf(stderr, "line = '%s'\n", line);
#endif
			break;
		}
		p++;
	}
	if ((*q = malloc(strlen(line)+1)) == NULL) {
		fprintf(stderr, "unable to malloc %d bytes: %s\n",
			strlen(line)+1, strerror(errno));
		exit(1);
	}
	strcpy(*q, line);
}
